import logging
from math import floor
from typing import List, Optional, Tuple

from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

logger = logging.getLogger(__name__)


class SICDUpdater:
    """
    This class provides a means to perform common updates to a SICD XML metadata document.
    """

    def __init__(self, xml_str: str):
        """
        Construct a new instance of this class to manage a given set of SICD metadata.

        :param xml_str: the SICD XML metadata to update
        """
        self.xml_str = xml_str
        if self.xml_str is not None and len(self.xml_str) > 0:
            parser = XmlParser()
            self.sicd = parser.from_string(self.xml_str)

        # Here we're storing off the original first row/col to support the case where multiple chips are
        # created from a SICD image that has already been chipped.
        self.original_first_row = self.sicd.image_data.first_row
        self.original_first_col = self.sicd.image_data.first_col

    def update_image_data_for_chip(self, chip_bounds: List[int], output_size: Optional[Tuple[int, int]]) -> None:
        """
        This updates the SICD ImageData structure so that the FirstRow, FirstCol and NumRows, NumCols
        elements match the new chip boundary.A sample of this XML structure is shown below::

            <ImageData>
                <NumRows>3859</NumRows>
                <NumCols>6679</NumCols>
                <FirstRow>0</FirstRow>
                <FirstCol>0</FirstCol>
                <FullImage>
                    <NumRows>3859</NumRows>
                    <NumCols>6679</NumCols>
                </FullImage>
            </ImageData>


        :param chip_bounds: the [col, row, width, height] of the chip boundary
        :param output_size: the [width, height] of the output chip
        """

        if output_size is not None and (output_size[0] != chip_bounds[2] or output_size[1] != chip_bounds[3]):
            raise ValueError("SICD chipping does not support scaling operations.")

        self.sicd.image_data.first_row = floor(float(self.original_first_row)) + int(chip_bounds[1])
        self.sicd.image_data.first_col = floor(float(self.original_first_col)) + int(chip_bounds[0])
        self.sicd.image_data.num_rows = int(chip_bounds[3])
        self.sicd.image_data.num_cols = int(chip_bounds[2])

    def encode_current_xml(self) -> str:
        """
        Returns a copy of the current SICD metadata encoded in XML.

        :return: xml encoded SICD metadata
        """
        serializer = XmlSerializer(config=SerializerConfig(pretty_print=False))
        updated_xml = serializer.render(self.sicd)
        return updated_xml



# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa
"""
The image_processing package contains various utilities for manipulating overhead imagery.

Image Tiling: Tiling With Updated Image Metadata
************************************************

Many applications break large remote sensing images into smaller chips or tiles for distributed processing or
dissemination. GDAL's Translate function provides basic capabilities, but it does not correctly update geospatial
metadata to reflect the new image extent. These utilities provide those functions so tile consumers can correctly
interpret the pixel information they have been provided.

.. code-block:: python
    :caption: Example showing creation of a NITF tile from the upper left corner of an image

    # Load the image and create a sensor model
    ds, sensor_model = load_gdal_dataset("./imagery/sample.nitf")
    tile_factory = GDALTileFactory(ds,
                                   sensor_model,
                                   GDALImageFormats.NITF,
                                   GDALCompressionOptions.NONE
                                   )

    # Bounds are [left_x, top_y, width, height]
    nitf_encoded_tile_bytes = tile_factory.create_encoded_tile([0, 0, 1024, 1024])


Image Tiling: Tiles for Display
*******************************

Some images, for example 11-bit panchromatic images or SAR imagery with floating point complex data, can not be
displayed directly without remapping the pixels into an 8-bit per pixel grayscale or RGB color model. The TileFactory
supports creation of tiles suitable for human review by setting both the output_type and range_adjustment options.
Note that the output_size parameter can be used to generate lower resolution tiles. This operation will make use of
GDAL generated overviews if they are available to the dataset.

.. code-block:: python
    :caption: Example showing creation of a PNG tile scaled down from the full resolution image

    viz_tile_factory = GDALTileFactory(ds,
                                       sensor_model,
                                       GDALImageFormats.PNG,
                                       GDALCompressionOptions.NONE,
                                       output_type=gdalconst.GDT_Byte,
                                       range_adjustment=RangeAdjustmentType.DRA)

    viz_tile = viz_tile_factory.create_encoded_tile([0, 0, 1024, 1024], output_size=(512, 512))


Complex SAR Data Display
************************

There are a variety of different techniques to convert complex SAR data to a simple image suitable for human display.
The toolkit contains two helper functions that can convert complex image data into an 8-bit grayscle representation
The equations implemented are described in Sections 3.1 and 3.2 of SAR Image Scaling, Dynamic Range, Radiometric
Calibration, and Display (SAND2019-2371).

.. code-block:: python
    :caption: Example converting complex SAR data into a 8-bit per pixel image for display

    import numpy as np
    from aws.osml.image_processing import histogram_stretch, quarter_power_image

    sicd_dataset, sensor_model = load_gdal_dataset("./sample-sicd.nitf")
    complex_pixels = sicd_dataset.ReadAsArray()

    histo_stretch_pixels = histogram_stretch(complex_pixels)
    quarter_power_pixels = quarter_power_image(complex_pixels)


.. figure:: ../images/SAR-HistogramStretchImage.png
   :width: 400
   :alt: Histogram Stretch Applied to Sample SICD Image

    Example of applying histogram_stretch to a sample SICD image.


.. figure:: ../images/SAR-QuarterPowerImage.png
   :width: 400
   :alt: Quarter Power Image Applied to Sample SICD Image

    Example of applying quarter_power_image to a sample SICD image.


-------------------------

APIs
****
"""

from .gdal_tile_factory import GDALTileFactory
from .sar_complex_imageop import histogram_stretch, quarter_power_image

__all__ = ["GDALTileFactory", "histogram_stretch", "quarter_power_image"]



from .feature_index import Feature2DSpatialIndex, STRFeature2DSpatialIndex
from .imaged_feature_property_accessor import ImagedFeaturePropertyAccessor

"""
The features package contains classes that assist with working with geospatial features derived from imagery.
"""

__all__ = [
    "Feature2DSpatialIndex",
    "ImagedFeaturePropertyAccessor",
    "STRFeature2DSpatialIndex",
]
